# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gnes.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='gnes.proto',
  package='gnes',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\ngnes.proto\x12\x04gnes\x1a\x1fgoogle/protobuf/timestamp.proto\"9\n\x07NdArray\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\x11\n\x05shape\x18\x02 \x03(\rB\x02\x10\x01\x12\r\n\x05\x64type\x18\x03 \x01(\t\"\xbc\x01\n\x05\x43hunk\x12\x0e\n\x06\x64oc_id\x18\x01 \x01(\x04\x12\x0e\n\x04text\x18\x02 \x01(\tH\x00\x12\x1d\n\x04\x62lob\x18\x03 \x01(\x0b\x32\r.gnes.NdArrayH\x00\x12\x11\n\toffset_1d\x18\x04 \x01(\r\x12)\n\toffset_nd\x18\x05 \x01(\x0b\x32\x16.gnes.Chunk.Coordinate\x12\x0e\n\x06weight\x18\x06 \x01(\x02\x1a\x1b\n\nCoordinate\x12\r\n\x01x\x18\x01 \x03(\rB\x02\x10\x01\x42\t\n\x07\x63ontent\"\xed\x02\n\x08\x44ocument\x12\x0e\n\x06\x64oc_id\x18\x01 \x01(\x04\x12\x1b\n\x06\x63hunks\x18\x02 \x03(\x0b\x32\x0b.gnes.Chunk\x12\'\n\x10\x63hunk_embeddings\x18\x03 \x01(\x0b\x32\r.gnes.NdArray\x12(\n\x08\x64oc_type\x18\x04 \x01(\x0e\x32\x16.gnes.Document.DocType\x12\x11\n\tmeta_info\x18\x05 \x01(\x0c\x12\x12\n\x08raw_text\x18\x06 \x01(\tH\x00\x12\"\n\traw_image\x18\x07 \x01(\x0b\x32\r.gnes.NdArrayH\x00\x12\"\n\traw_video\x18\x08 \x01(\x0b\x32\r.gnes.NdArrayH\x00\x12\x13\n\traw_bytes\x18\t \x01(\x0cH\x00\x12\x0e\n\x06weight\x18\n \x01(\x02\"A\n\x07\x44ocType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x08\n\x04TEXT\x10\x01\x12\t\n\x05IMAGE\x10\x02\x12\t\n\x05VIDEO\x10\x03\x12\t\n\x05\x41UDIO\x10\x04\x42\n\n\x08raw_data\"\xd4\x01\n\x08\x45nvelope\x12\x11\n\tclient_id\x18\x01 \x01(\t\x12\x12\n\nrequest_id\x18\x02 \x01(\t\x12\x0f\n\x07part_id\x18\x03 \x01(\r\x12\x10\n\x08num_part\x18\x04 \x03(\r\x12\x0f\n\x07timeout\x18\x05 \x01(\r\x12$\n\x06routes\x18\x06 \x03(\x0b\x32\x14.gnes.Envelope.route\x1aG\n\x05route\x12\x0f\n\x07service\x18\x01 \x01(\t\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"y\n\x07Message\x12 \n\x08\x65nvelope\x18\x01 \x01(\x0b\x32\x0e.gnes.Envelope\x12 \n\x07request\x18\x02 \x01(\x0b\x32\r.gnes.RequestH\x00\x12\"\n\x08response\x18\x03 \x01(\x0b\x32\x0e.gnes.ResponseH\x00\x42\x06\n\x04\x62ody\"\xf6\x03\n\x07Request\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12+\n\x05train\x18\x02 \x01(\x0b\x32\x1a.gnes.Request.TrainRequestH\x00\x12+\n\x05index\x18\x03 \x01(\x0b\x32\x1a.gnes.Request.IndexRequestH\x00\x12,\n\x06search\x18\x04 \x01(\x0b\x32\x1a.gnes.Request.QueryRequestH\x00\x12/\n\x07\x63ontrol\x18\x05 \x01(\x0b\x32\x1c.gnes.Request.ControlRequestH\x00\x1a;\n\x0cTrainRequest\x12\x1c\n\x04\x64ocs\x18\x01 \x03(\x0b\x32\x0e.gnes.Document\x12\r\n\x05\x66lush\x18\x02 \x01(\x08\x1a,\n\x0cIndexRequest\x12\x1c\n\x04\x64ocs\x18\x01 \x03(\x0b\x32\x0e.gnes.Document\x1a<\n\x0cQueryRequest\x12\x1d\n\x05query\x18\x01 \x01(\x0b\x32\x0e.gnes.Document\x12\r\n\x05top_k\x18\x02 \x01(\r\x1am\n\x0e\x43ontrolRequest\x12\x35\n\x07\x63ommand\x18\x01 \x01(\x0e\x32$.gnes.Request.ControlRequest.Command\"$\n\x07\x43ommand\x12\r\n\tTERMINATE\x10\x00\x12\n\n\x06STATUS\x10\x01\x42\x06\n\x04\x62ody\"\xc4\x06\n\x08Response\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12-\n\x05train\x18\x02 \x01(\x0b\x32\x1c.gnes.Response.TrainResponseH\x00\x12-\n\x05index\x18\x03 \x01(\x0b\x32\x1c.gnes.Response.IndexResponseH\x00\x12.\n\x06search\x18\x04 \x01(\x0b\x32\x1c.gnes.Response.QueryResponseH\x00\x12\x31\n\x07\x63ontrol\x18\x05 \x01(\x0b\x32\x1e.gnes.Response.ControlResponseH\x00\x1a\x36\n\rTrainResponse\x12%\n\x06status\x18\x01 \x01(\x0e\x32\x15.gnes.Response.Status\x1a\x36\n\rIndexResponse\x12%\n\x06status\x18\x01 \x01(\x0e\x32\x15.gnes.Response.Status\x1a\x38\n\x0f\x43ontrolResponse\x12%\n\x06status\x18\x01 \x01(\x0e\x32\x15.gnes.Response.Status\x1a\x81\x03\n\rQueryResponse\x12%\n\x06status\x18\x01 \x01(\x0e\x32\x15.gnes.Response.Status\x12\r\n\x05top_k\x18\x02 \x01(\r\x12?\n\x0ctopk_results\x18\x03 \x03(\x0b\x32).gnes.Response.QueryResponse.ScoredResult\x12\x39\n\x05level\x18\x04 \x01(\x0e\x32*.gnes.Response.QueryResponse.ResponseLevel\x1a{\n\x0cScoredResult\x12\x1c\n\x05\x63hunk\x18\x01 \x01(\x0b\x32\x0b.gnes.ChunkH\x00\x12\x1d\n\x03\x64oc\x18\x02 \x01(\x0b\x32\x0e.gnes.DocumentH\x00\x12\r\n\x05score\x18\x03 \x01(\x02\x12\x17\n\x0fscore_explained\x18\x04 \x01(\tB\x06\n\x04\x62ody\"A\n\rResponseLevel\x12\t\n\x05\x43HUNK\x10\x00\x12\x17\n\x13\x44OCUMENT_NOT_FILLED\x10\x01\x12\x0c\n\x08\x44OCUMENT\x10\x02\"-\n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\x12\x0b\n\x07PENDING\x10\x02\x42\x06\n\x04\x62ody2\xe3\x01\n\x07GnesRPC\x12(\n\x05Train\x12\r.gnes.Request\x1a\x0e.gnes.Response\"\x00\x12(\n\x05Index\x12\r.gnes.Request\x1a\x0e.gnes.Response\"\x00\x12(\n\x05Query\x12\r.gnes.Request\x1a\x0e.gnes.Response\"\x00\x12\'\n\x04\x43\x61ll\x12\r.gnes.Request\x1a\x0e.gnes.Response\"\x00\x12\x31\n\nStreamCall\x12\r.gnes.Request\x1a\x0e.gnes.Response\"\x00(\x01\x30\x01\x62\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])



_DOCUMENT_DOCTYPE = _descriptor.EnumDescriptor(
  name='DocType',
  full_name='gnes.Document.DocType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEXT', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IMAGE', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='VIDEO', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AUDIO', index=4, number=4,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=592,
  serialized_end=657,
)
_sym_db.RegisterEnumDescriptor(_DOCUMENT_DOCTYPE)

_REQUEST_CONTROLREQUEST_COMMAND = _descriptor.EnumDescriptor(
  name='Command',
  full_name='gnes.Request.ControlRequest.Command',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TERMINATE', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STATUS', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1468,
  serialized_end=1504,
)
_sym_db.RegisterEnumDescriptor(_REQUEST_CONTROLREQUEST_COMMAND)

_RESPONSE_QUERYRESPONSE_RESPONSELEVEL = _descriptor.EnumDescriptor(
  name='ResponseLevel',
  full_name='gnes.Response.QueryResponse.ResponseLevel',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CHUNK', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DOCUMENT_NOT_FILLED', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DOCUMENT', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2231,
  serialized_end=2296,
)
_sym_db.RegisterEnumDescriptor(_RESPONSE_QUERYRESPONSE_RESPONSELEVEL)

_RESPONSE_STATUS = _descriptor.EnumDescriptor(
  name='Status',
  full_name='gnes.Response.Status',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SUCCESS', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PENDING', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2298,
  serialized_end=2343,
)
_sym_db.RegisterEnumDescriptor(_RESPONSE_STATUS)


_NDARRAY = _descriptor.Descriptor(
  name='NdArray',
  full_name='gnes.NdArray',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='gnes.NdArray.data', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shape', full_name='gnes.NdArray.shape', index=1,
      number=2, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dtype', full_name='gnes.NdArray.dtype', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=53,
  serialized_end=110,
)


_CHUNK_COORDINATE = _descriptor.Descriptor(
  name='Coordinate',
  full_name='gnes.Chunk.Coordinate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='gnes.Chunk.Coordinate.x', index=0,
      number=1, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=263,
  serialized_end=290,
)

_CHUNK = _descriptor.Descriptor(
  name='Chunk',
  full_name='gnes.Chunk',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='doc_id', full_name='gnes.Chunk.doc_id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='text', full_name='gnes.Chunk.text', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blob', full_name='gnes.Chunk.blob', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='offset_1d', full_name='gnes.Chunk.offset_1d', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='offset_nd', full_name='gnes.Chunk.offset_nd', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weight', full_name='gnes.Chunk.weight', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_CHUNK_COORDINATE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='content', full_name='gnes.Chunk.content',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=113,
  serialized_end=301,
)


_DOCUMENT = _descriptor.Descriptor(
  name='Document',
  full_name='gnes.Document',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='doc_id', full_name='gnes.Document.doc_id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='chunks', full_name='gnes.Document.chunks', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='chunk_embeddings', full_name='gnes.Document.chunk_embeddings', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='doc_type', full_name='gnes.Document.doc_type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='meta_info', full_name='gnes.Document.meta_info', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='raw_text', full_name='gnes.Document.raw_text', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='raw_image', full_name='gnes.Document.raw_image', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='raw_video', full_name='gnes.Document.raw_video', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='raw_bytes', full_name='gnes.Document.raw_bytes', index=8,
      number=9, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weight', full_name='gnes.Document.weight', index=9,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _DOCUMENT_DOCTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='raw_data', full_name='gnes.Document.raw_data',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=304,
  serialized_end=669,
)


_ENVELOPE_ROUTE = _descriptor.Descriptor(
  name='route',
  full_name='gnes.Envelope.route',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='service', full_name='gnes.Envelope.route.service', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='gnes.Envelope.route.timestamp', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=813,
  serialized_end=884,
)

_ENVELOPE = _descriptor.Descriptor(
  name='Envelope',
  full_name='gnes.Envelope',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='client_id', full_name='gnes.Envelope.client_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='request_id', full_name='gnes.Envelope.request_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='part_id', full_name='gnes.Envelope.part_id', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='num_part', full_name='gnes.Envelope.num_part', index=3,
      number=4, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timeout', full_name='gnes.Envelope.timeout', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='routes', full_name='gnes.Envelope.routes', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_ENVELOPE_ROUTE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=672,
  serialized_end=884,
)


_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='gnes.Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='envelope', full_name='gnes.Message.envelope', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='request', full_name='gnes.Message.request', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='response', full_name='gnes.Message.response', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='body', full_name='gnes.Message.body',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=886,
  serialized_end=1007,
)


_REQUEST_TRAINREQUEST = _descriptor.Descriptor(
  name='TrainRequest',
  full_name='gnes.Request.TrainRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='docs', full_name='gnes.Request.TrainRequest.docs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='flush', full_name='gnes.Request.TrainRequest.flush', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1226,
  serialized_end=1285,
)

_REQUEST_INDEXREQUEST = _descriptor.Descriptor(
  name='IndexRequest',
  full_name='gnes.Request.IndexRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='docs', full_name='gnes.Request.IndexRequest.docs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1287,
  serialized_end=1331,
)

_REQUEST_QUERYREQUEST = _descriptor.Descriptor(
  name='QueryRequest',
  full_name='gnes.Request.QueryRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='query', full_name='gnes.Request.QueryRequest.query', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='top_k', full_name='gnes.Request.QueryRequest.top_k', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1333,
  serialized_end=1393,
)

_REQUEST_CONTROLREQUEST = _descriptor.Descriptor(
  name='ControlRequest',
  full_name='gnes.Request.ControlRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='command', full_name='gnes.Request.ControlRequest.command', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _REQUEST_CONTROLREQUEST_COMMAND,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1395,
  serialized_end=1504,
)

_REQUEST = _descriptor.Descriptor(
  name='Request',
  full_name='gnes.Request',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_id', full_name='gnes.Request.request_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='train', full_name='gnes.Request.train', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='index', full_name='gnes.Request.index', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='search', full_name='gnes.Request.search', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='control', full_name='gnes.Request.control', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_REQUEST_TRAINREQUEST, _REQUEST_INDEXREQUEST, _REQUEST_QUERYREQUEST, _REQUEST_CONTROLREQUEST, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='body', full_name='gnes.Request.body',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=1010,
  serialized_end=1512,
)


_RESPONSE_TRAINRESPONSE = _descriptor.Descriptor(
  name='TrainResponse',
  full_name='gnes.Response.TrainResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='gnes.Response.TrainResponse.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1740,
  serialized_end=1794,
)

_RESPONSE_INDEXRESPONSE = _descriptor.Descriptor(
  name='IndexResponse',
  full_name='gnes.Response.IndexResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='gnes.Response.IndexResponse.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1796,
  serialized_end=1850,
)

_RESPONSE_CONTROLRESPONSE = _descriptor.Descriptor(
  name='ControlResponse',
  full_name='gnes.Response.ControlResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='gnes.Response.ControlResponse.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1852,
  serialized_end=1908,
)

_RESPONSE_QUERYRESPONSE_SCOREDRESULT = _descriptor.Descriptor(
  name='ScoredResult',
  full_name='gnes.Response.QueryResponse.ScoredResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='chunk', full_name='gnes.Response.QueryResponse.ScoredResult.chunk', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='doc', full_name='gnes.Response.QueryResponse.ScoredResult.doc', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='score', full_name='gnes.Response.QueryResponse.ScoredResult.score', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='score_explained', full_name='gnes.Response.QueryResponse.ScoredResult.score_explained', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='body', full_name='gnes.Response.QueryResponse.ScoredResult.body',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=2106,
  serialized_end=2229,
)

_RESPONSE_QUERYRESPONSE = _descriptor.Descriptor(
  name='QueryResponse',
  full_name='gnes.Response.QueryResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='gnes.Response.QueryResponse.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='top_k', full_name='gnes.Response.QueryResponse.top_k', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='topk_results', full_name='gnes.Response.QueryResponse.topk_results', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='level', full_name='gnes.Response.QueryResponse.level', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_RESPONSE_QUERYRESPONSE_SCOREDRESULT, ],
  enum_types=[
    _RESPONSE_QUERYRESPONSE_RESPONSELEVEL,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1911,
  serialized_end=2296,
)

_RESPONSE = _descriptor.Descriptor(
  name='Response',
  full_name='gnes.Response',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_id', full_name='gnes.Response.request_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='train', full_name='gnes.Response.train', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='index', full_name='gnes.Response.index', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='search', full_name='gnes.Response.search', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='control', full_name='gnes.Response.control', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_RESPONSE_TRAINRESPONSE, _RESPONSE_INDEXRESPONSE, _RESPONSE_CONTROLRESPONSE, _RESPONSE_QUERYRESPONSE, ],
  enum_types=[
    _RESPONSE_STATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='body', full_name='gnes.Response.body',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=1515,
  serialized_end=2351,
)

_CHUNK_COORDINATE.containing_type = _CHUNK
_CHUNK.fields_by_name['blob'].message_type = _NDARRAY
_CHUNK.fields_by_name['offset_nd'].message_type = _CHUNK_COORDINATE
_CHUNK.oneofs_by_name['content'].fields.append(
  _CHUNK.fields_by_name['text'])
_CHUNK.fields_by_name['text'].containing_oneof = _CHUNK.oneofs_by_name['content']
_CHUNK.oneofs_by_name['content'].fields.append(
  _CHUNK.fields_by_name['blob'])
_CHUNK.fields_by_name['blob'].containing_oneof = _CHUNK.oneofs_by_name['content']
_DOCUMENT.fields_by_name['chunks'].message_type = _CHUNK
_DOCUMENT.fields_by_name['chunk_embeddings'].message_type = _NDARRAY
_DOCUMENT.fields_by_name['doc_type'].enum_type = _DOCUMENT_DOCTYPE
_DOCUMENT.fields_by_name['raw_image'].message_type = _NDARRAY
_DOCUMENT.fields_by_name['raw_video'].message_type = _NDARRAY
_DOCUMENT_DOCTYPE.containing_type = _DOCUMENT
_DOCUMENT.oneofs_by_name['raw_data'].fields.append(
  _DOCUMENT.fields_by_name['raw_text'])
_DOCUMENT.fields_by_name['raw_text'].containing_oneof = _DOCUMENT.oneofs_by_name['raw_data']
_DOCUMENT.oneofs_by_name['raw_data'].fields.append(
  _DOCUMENT.fields_by_name['raw_image'])
_DOCUMENT.fields_by_name['raw_image'].containing_oneof = _DOCUMENT.oneofs_by_name['raw_data']
_DOCUMENT.oneofs_by_name['raw_data'].fields.append(
  _DOCUMENT.fields_by_name['raw_video'])
_DOCUMENT.fields_by_name['raw_video'].containing_oneof = _DOCUMENT.oneofs_by_name['raw_data']
_DOCUMENT.oneofs_by_name['raw_data'].fields.append(
  _DOCUMENT.fields_by_name['raw_bytes'])
_DOCUMENT.fields_by_name['raw_bytes'].containing_oneof = _DOCUMENT.oneofs_by_name['raw_data']
_ENVELOPE_ROUTE.fields_by_name['timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ENVELOPE_ROUTE.containing_type = _ENVELOPE
_ENVELOPE.fields_by_name['routes'].message_type = _ENVELOPE_ROUTE
_MESSAGE.fields_by_name['envelope'].message_type = _ENVELOPE
_MESSAGE.fields_by_name['request'].message_type = _REQUEST
_MESSAGE.fields_by_name['response'].message_type = _RESPONSE
_MESSAGE.oneofs_by_name['body'].fields.append(
  _MESSAGE.fields_by_name['request'])
_MESSAGE.fields_by_name['request'].containing_oneof = _MESSAGE.oneofs_by_name['body']
_MESSAGE.oneofs_by_name['body'].fields.append(
  _MESSAGE.fields_by_name['response'])
_MESSAGE.fields_by_name['response'].containing_oneof = _MESSAGE.oneofs_by_name['body']
_REQUEST_TRAINREQUEST.fields_by_name['docs'].message_type = _DOCUMENT
_REQUEST_TRAINREQUEST.containing_type = _REQUEST
_REQUEST_INDEXREQUEST.fields_by_name['docs'].message_type = _DOCUMENT
_REQUEST_INDEXREQUEST.containing_type = _REQUEST
_REQUEST_QUERYREQUEST.fields_by_name['query'].message_type = _DOCUMENT
_REQUEST_QUERYREQUEST.containing_type = _REQUEST
_REQUEST_CONTROLREQUEST.fields_by_name['command'].enum_type = _REQUEST_CONTROLREQUEST_COMMAND
_REQUEST_CONTROLREQUEST.containing_type = _REQUEST
_REQUEST_CONTROLREQUEST_COMMAND.containing_type = _REQUEST_CONTROLREQUEST
_REQUEST.fields_by_name['train'].message_type = _REQUEST_TRAINREQUEST
_REQUEST.fields_by_name['index'].message_type = _REQUEST_INDEXREQUEST
_REQUEST.fields_by_name['search'].message_type = _REQUEST_QUERYREQUEST
_REQUEST.fields_by_name['control'].message_type = _REQUEST_CONTROLREQUEST
_REQUEST.oneofs_by_name['body'].fields.append(
  _REQUEST.fields_by_name['train'])
_REQUEST.fields_by_name['train'].containing_oneof = _REQUEST.oneofs_by_name['body']
_REQUEST.oneofs_by_name['body'].fields.append(
  _REQUEST.fields_by_name['index'])
_REQUEST.fields_by_name['index'].containing_oneof = _REQUEST.oneofs_by_name['body']
_REQUEST.oneofs_by_name['body'].fields.append(
  _REQUEST.fields_by_name['search'])
_REQUEST.fields_by_name['search'].containing_oneof = _REQUEST.oneofs_by_name['body']
_REQUEST.oneofs_by_name['body'].fields.append(
  _REQUEST.fields_by_name['control'])
_REQUEST.fields_by_name['control'].containing_oneof = _REQUEST.oneofs_by_name['body']
_RESPONSE_TRAINRESPONSE.fields_by_name['status'].enum_type = _RESPONSE_STATUS
_RESPONSE_TRAINRESPONSE.containing_type = _RESPONSE
_RESPONSE_INDEXRESPONSE.fields_by_name['status'].enum_type = _RESPONSE_STATUS
_RESPONSE_INDEXRESPONSE.containing_type = _RESPONSE
_RESPONSE_CONTROLRESPONSE.fields_by_name['status'].enum_type = _RESPONSE_STATUS
_RESPONSE_CONTROLRESPONSE.containing_type = _RESPONSE
_RESPONSE_QUERYRESPONSE_SCOREDRESULT.fields_by_name['chunk'].message_type = _CHUNK
_RESPONSE_QUERYRESPONSE_SCOREDRESULT.fields_by_name['doc'].message_type = _DOCUMENT
_RESPONSE_QUERYRESPONSE_SCOREDRESULT.containing_type = _RESPONSE_QUERYRESPONSE
_RESPONSE_QUERYRESPONSE_SCOREDRESULT.oneofs_by_name['body'].fields.append(
  _RESPONSE_QUERYRESPONSE_SCOREDRESULT.fields_by_name['chunk'])
_RESPONSE_QUERYRESPONSE_SCOREDRESULT.fields_by_name['chunk'].containing_oneof = _RESPONSE_QUERYRESPONSE_SCOREDRESULT.oneofs_by_name['body']
_RESPONSE_QUERYRESPONSE_SCOREDRESULT.oneofs_by_name['body'].fields.append(
  _RESPONSE_QUERYRESPONSE_SCOREDRESULT.fields_by_name['doc'])
_RESPONSE_QUERYRESPONSE_SCOREDRESULT.fields_by_name['doc'].containing_oneof = _RESPONSE_QUERYRESPONSE_SCOREDRESULT.oneofs_by_name['body']
_RESPONSE_QUERYRESPONSE.fields_by_name['status'].enum_type = _RESPONSE_STATUS
_RESPONSE_QUERYRESPONSE.fields_by_name['topk_results'].message_type = _RESPONSE_QUERYRESPONSE_SCOREDRESULT
_RESPONSE_QUERYRESPONSE.fields_by_name['level'].enum_type = _RESPONSE_QUERYRESPONSE_RESPONSELEVEL
_RESPONSE_QUERYRESPONSE.containing_type = _RESPONSE
_RESPONSE_QUERYRESPONSE_RESPONSELEVEL.containing_type = _RESPONSE_QUERYRESPONSE
_RESPONSE.fields_by_name['train'].message_type = _RESPONSE_TRAINRESPONSE
_RESPONSE.fields_by_name['index'].message_type = _RESPONSE_INDEXRESPONSE
_RESPONSE.fields_by_name['search'].message_type = _RESPONSE_QUERYRESPONSE
_RESPONSE.fields_by_name['control'].message_type = _RESPONSE_CONTROLRESPONSE
_RESPONSE_STATUS.containing_type = _RESPONSE
_RESPONSE.oneofs_by_name['body'].fields.append(
  _RESPONSE.fields_by_name['train'])
_RESPONSE.fields_by_name['train'].containing_oneof = _RESPONSE.oneofs_by_name['body']
_RESPONSE.oneofs_by_name['body'].fields.append(
  _RESPONSE.fields_by_name['index'])
_RESPONSE.fields_by_name['index'].containing_oneof = _RESPONSE.oneofs_by_name['body']
_RESPONSE.oneofs_by_name['body'].fields.append(
  _RESPONSE.fields_by_name['search'])
_RESPONSE.fields_by_name['search'].containing_oneof = _RESPONSE.oneofs_by_name['body']
_RESPONSE.oneofs_by_name['body'].fields.append(
  _RESPONSE.fields_by_name['control'])
_RESPONSE.fields_by_name['control'].containing_oneof = _RESPONSE.oneofs_by_name['body']
DESCRIPTOR.message_types_by_name['NdArray'] = _NDARRAY
DESCRIPTOR.message_types_by_name['Chunk'] = _CHUNK
DESCRIPTOR.message_types_by_name['Document'] = _DOCUMENT
DESCRIPTOR.message_types_by_name['Envelope'] = _ENVELOPE
DESCRIPTOR.message_types_by_name['Message'] = _MESSAGE
DESCRIPTOR.message_types_by_name['Request'] = _REQUEST
DESCRIPTOR.message_types_by_name['Response'] = _RESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

NdArray = _reflection.GeneratedProtocolMessageType('NdArray', (_message.Message,), {
  'DESCRIPTOR' : _NDARRAY,
  '__module__' : 'gnes_pb2'
  # @@protoc_insertion_point(class_scope:gnes.NdArray)
  })
_sym_db.RegisterMessage(NdArray)

Chunk = _reflection.GeneratedProtocolMessageType('Chunk', (_message.Message,), {

  'Coordinate' : _reflection.GeneratedProtocolMessageType('Coordinate', (_message.Message,), {
    'DESCRIPTOR' : _CHUNK_COORDINATE,
    '__module__' : 'gnes_pb2'
    # @@protoc_insertion_point(class_scope:gnes.Chunk.Coordinate)
    })
  ,
  'DESCRIPTOR' : _CHUNK,
  '__module__' : 'gnes_pb2'
  # @@protoc_insertion_point(class_scope:gnes.Chunk)
  })
_sym_db.RegisterMessage(Chunk)
_sym_db.RegisterMessage(Chunk.Coordinate)

Document = _reflection.GeneratedProtocolMessageType('Document', (_message.Message,), {
  'DESCRIPTOR' : _DOCUMENT,
  '__module__' : 'gnes_pb2'
  # @@protoc_insertion_point(class_scope:gnes.Document)
  })
_sym_db.RegisterMessage(Document)

Envelope = _reflection.GeneratedProtocolMessageType('Envelope', (_message.Message,), {

  'route' : _reflection.GeneratedProtocolMessageType('route', (_message.Message,), {
    'DESCRIPTOR' : _ENVELOPE_ROUTE,
    '__module__' : 'gnes_pb2'
    # @@protoc_insertion_point(class_scope:gnes.Envelope.route)
    })
  ,
  'DESCRIPTOR' : _ENVELOPE,
  '__module__' : 'gnes_pb2'
  # @@protoc_insertion_point(class_scope:gnes.Envelope)
  })
_sym_db.RegisterMessage(Envelope)
_sym_db.RegisterMessage(Envelope.route)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGE,
  '__module__' : 'gnes_pb2'
  # @@protoc_insertion_point(class_scope:gnes.Message)
  })
_sym_db.RegisterMessage(Message)

Request = _reflection.GeneratedProtocolMessageType('Request', (_message.Message,), {

  'TrainRequest' : _reflection.GeneratedProtocolMessageType('TrainRequest', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_TRAINREQUEST,
    '__module__' : 'gnes_pb2'
    # @@protoc_insertion_point(class_scope:gnes.Request.TrainRequest)
    })
  ,

  'IndexRequest' : _reflection.GeneratedProtocolMessageType('IndexRequest', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_INDEXREQUEST,
    '__module__' : 'gnes_pb2'
    # @@protoc_insertion_point(class_scope:gnes.Request.IndexRequest)
    })
  ,

  'QueryRequest' : _reflection.GeneratedProtocolMessageType('QueryRequest', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_QUERYREQUEST,
    '__module__' : 'gnes_pb2'
    # @@protoc_insertion_point(class_scope:gnes.Request.QueryRequest)
    })
  ,

  'ControlRequest' : _reflection.GeneratedProtocolMessageType('ControlRequest', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_CONTROLREQUEST,
    '__module__' : 'gnes_pb2'
    # @@protoc_insertion_point(class_scope:gnes.Request.ControlRequest)
    })
  ,
  'DESCRIPTOR' : _REQUEST,
  '__module__' : 'gnes_pb2'
  # @@protoc_insertion_point(class_scope:gnes.Request)
  })
_sym_db.RegisterMessage(Request)
_sym_db.RegisterMessage(Request.TrainRequest)
_sym_db.RegisterMessage(Request.IndexRequest)
_sym_db.RegisterMessage(Request.QueryRequest)
_sym_db.RegisterMessage(Request.ControlRequest)

Response = _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {

  'TrainResponse' : _reflection.GeneratedProtocolMessageType('TrainResponse', (_message.Message,), {
    'DESCRIPTOR' : _RESPONSE_TRAINRESPONSE,
    '__module__' : 'gnes_pb2'
    # @@protoc_insertion_point(class_scope:gnes.Response.TrainResponse)
    })
  ,

  'IndexResponse' : _reflection.GeneratedProtocolMessageType('IndexResponse', (_message.Message,), {
    'DESCRIPTOR' : _RESPONSE_INDEXRESPONSE,
    '__module__' : 'gnes_pb2'
    # @@protoc_insertion_point(class_scope:gnes.Response.IndexResponse)
    })
  ,

  'ControlResponse' : _reflection.GeneratedProtocolMessageType('ControlResponse', (_message.Message,), {
    'DESCRIPTOR' : _RESPONSE_CONTROLRESPONSE,
    '__module__' : 'gnes_pb2'
    # @@protoc_insertion_point(class_scope:gnes.Response.ControlResponse)
    })
  ,

  'QueryResponse' : _reflection.GeneratedProtocolMessageType('QueryResponse', (_message.Message,), {

    'ScoredResult' : _reflection.GeneratedProtocolMessageType('ScoredResult', (_message.Message,), {
      'DESCRIPTOR' : _RESPONSE_QUERYRESPONSE_SCOREDRESULT,
      '__module__' : 'gnes_pb2'
      # @@protoc_insertion_point(class_scope:gnes.Response.QueryResponse.ScoredResult)
      })
    ,
    'DESCRIPTOR' : _RESPONSE_QUERYRESPONSE,
    '__module__' : 'gnes_pb2'
    # @@protoc_insertion_point(class_scope:gnes.Response.QueryResponse)
    })
  ,
  'DESCRIPTOR' : _RESPONSE,
  '__module__' : 'gnes_pb2'
  # @@protoc_insertion_point(class_scope:gnes.Response)
  })
_sym_db.RegisterMessage(Response)
_sym_db.RegisterMessage(Response.TrainResponse)
_sym_db.RegisterMessage(Response.IndexResponse)
_sym_db.RegisterMessage(Response.ControlResponse)
_sym_db.RegisterMessage(Response.QueryResponse)
_sym_db.RegisterMessage(Response.QueryResponse.ScoredResult)


_NDARRAY.fields_by_name['shape']._options = None
_CHUNK_COORDINATE.fields_by_name['x']._options = None

_GNESRPC = _descriptor.ServiceDescriptor(
  name='GnesRPC',
  full_name='gnes.GnesRPC',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=2354,
  serialized_end=2581,
  methods=[
  _descriptor.MethodDescriptor(
    name='Train',
    full_name='gnes.GnesRPC.Train',
    index=0,
    containing_service=None,
    input_type=_REQUEST,
    output_type=_RESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='Index',
    full_name='gnes.GnesRPC.Index',
    index=1,
    containing_service=None,
    input_type=_REQUEST,
    output_type=_RESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='Query',
    full_name='gnes.GnesRPC.Query',
    index=2,
    containing_service=None,
    input_type=_REQUEST,
    output_type=_RESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='Call',
    full_name='gnes.GnesRPC.Call',
    index=3,
    containing_service=None,
    input_type=_REQUEST,
    output_type=_RESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='StreamCall',
    full_name='gnes.GnesRPC.StreamCall',
    index=4,
    containing_service=None,
    input_type=_REQUEST,
    output_type=_RESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_GNESRPC)

DESCRIPTOR.services_by_name['GnesRPC'] = _GNESRPC

# @@protoc_insertion_point(module_scope)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


# A key-value map for Class to the (module)file it located in
from ..base import register_all_class

_cls2file_map = {
    'BasePreprocessor': 'base',
    'PipelinePreprocessor': 'base',
    'TextPreprocessor': 'text.simple',
    'BaseImagePreprocessor': 'image.base',
    'BaseTextPreprocessor': 'text.base',
    'BaseSlidingPreprocessor': 'image.sliding_window',
    'VanillaSlidingPreprocessor': 'image.sliding_window',
    'WeightedSlidingPreprocessor': 'image.sliding_window',
    'SegmentPreprocessor': 'image.segmentation',
    'BaseUnaryPreprocessor': 'base',
    'ResizeChunkPreprocessor': 'image.resize',
    'BaseVideoPreprocessor': 'video.base',
    'FFmpegPreprocessor': 'video.ffmpeg',
    'FFmpegVideoSegmentor': 'video.ffmpeg',
    'ShotDetectPreprocessor': 'video.shotdetect',
    'AudioVanilla': 'audio.audio_vanilla',
    'BaseAudioPreprocessor': 'audio.base'
}

register_all_class(_cls2file_map, 'preprocessor')



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.



from typing import List, Tuple, Any

import numpy as np

from ..base import BaseVectorIndexer
from ..key_only import ListKeyIndexer


class NumpyIndexer(BaseVectorIndexer):

    def __init__(self, num_bytes: int = None, *args, **kwargs):
        super().__init__()
        self.num_bytes = num_bytes
        self._vectors = None  # type: np.ndarray
        self._key_info_indexer = ListKeyIndexer()

    def add(self, keys: List[Tuple[int, Any]], vectors: np.ndarray, weights: List[float], *args,
            **kwargs):
        if len(vectors) % len(keys) != 0:
            raise ValueError('vectors bytes should be divided by doc_ids')

        if not self.num_bytes:
            self.num_bytes = vectors.shape[1]
        elif self.num_bytes != vectors.shape[1]:
            raise ValueError(
                "vectors' shape [%d, %d] does not match with indexer's dim: %d" %
                (vectors.shape[0], vectors.shape[1], self.num_bytes))

        if self._vectors is not None:
            self._vectors = np.concatenate([self._vectors, vectors], axis=0)
        else:
            self._vectors = vectors
        self._key_info_indexer.add(keys, weights)

    def query(self, keys: np.ndarray, top_k: int, *args, **kwargs
              ) -> List[List[Tuple]]:
        keys = np.expand_dims(keys, axis=1)
        dist = keys - np.expand_dims(self._vectors, axis=0)
        score = 1 - np.sum(np.minimum(np.abs(dist), 1), -1) / self.num_bytes

        ret = []
        for ids in score:
            rk = sorted(enumerate(ids), key=lambda x: -x[1])
            chunk_info = self._key_info_indexer.query([j[0] for j in rk])

            ret.append([(*r, s) for r, s in zip(chunk_info, [j[1] for j in rk])])
        return ret



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


# noinspection PyUnresolvedReferences

from typing import List

import numpy as np

from ..base import BaseTextEncoder
from ...helper import batching, pooling_torch


class GPTEncoder(BaseTextEncoder):
    is_trained = True
    batch_size = 64

    def __init__(self,
                 model_dir: str,
                 use_cuda: bool = False,
                 pooling_strategy: str = 'REDUCE_MEAN',
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self.model_dir = model_dir
        self.pooling_strategy = pooling_strategy
        self._use_cuda = use_cuda

    def post_init(self):
        import torch
        # Load pre-trained model tokenizer (vocabulary)
        self._init_model_tokenizer()
        self._use_cuda = self._use_cuda and torch.cuda.is_available()
        if self._use_cuda: self._model.to('cuda')

    def _get_token_ids(self, x):
        return self._tokenizer.convert_tokens_to_ids(self._tokenizer.tokenize(x))

    def _get_output_tensor(self, x):
        return self._model(x)

    def _init_model_tokenizer(self):
        from pytorch_pretrained_bert import OpenAIGPTTokenizer, OpenAIGPTModel
        self._tokenizer = OpenAIGPTTokenizer.from_pretrained(self.model_dir)
        self._model = OpenAIGPTModel.from_pretrained(self.model_dir)
        self._model.eval()

    @batching
    def encode(self, text: List[str], *args, **kwargs) -> np.ndarray:
        import torch
        batch_size = len(text)

        # tokenize text
        tokens_ids = []
        tokens_lens = []
        max_len = 0
        for _ in text:
            # Convert token to vocabulary indices
            token_ids = self._get_token_ids(_)
            token_len = len(token_ids)

            if max_len < token_len:
                max_len = token_len

            tokens_ids.append(token_ids)
            tokens_lens.append(token_len)

        batch_data = np.zeros([batch_size, max_len], dtype=np.int64)
        # batch_mask = np.zeros([batch_size, max_len], dtype=np.float32)
        for i, ids in enumerate(tokens_ids):
            batch_data[i, :tokens_lens[i]] = tokens_ids[i]
            # batch_mask[i, :tokens_lens[i]] = 1

        # Convert inputs to PyTorch tensors
        tokens_tensor = torch.tensor(batch_data)
        tokens_lens = torch.LongTensor(tokens_lens)
        mask_tensor = torch.arange(max_len)[None, :] < tokens_lens[:, None]
        mask_tensor = mask_tensor.to(
            mask_tensor.device, dtype=torch.float32)

        if self._use_cuda:
            # If you have a GPU, put everything on cuda
            tokens_tensor = tokens_tensor.to('cuda')
            mask_tensor = mask_tensor.to('cuda')

        # Predict hidden states features for each layer
        with torch.no_grad():
            # the encoded-hidden-states at the top of the model, as a
            # torch.FloatTensor of size [batch_size, sequence_length, hidden_size]
            output_tensor = self._get_output_tensor(tokens_tensor)
            output_tensor = pooling_torch(output_tensor, mask_tensor, self.pooling_strategy)

        if self._use_cuda:
            output_tensor = output_tensor.cpu()
        return output_tensor.numpy()


class GPT2Encoder(GPTEncoder):

    def _get_token_ids(self, x):
        return self._tokenizer.encode(x)

    def _get_output_tensor(self, x):
        return self._model(x)[0]

    def _init_model_tokenizer(self):
        from pytorch_pretrained_bert import GPT2Model, GPT2Tokenizer
        self._tokenizer = GPT2Tokenizer.from_pretrained(self.model_dir)
        self._model = GPT2Model.from_pretrained(self.model_dir)
        self._model.eval()



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

from ..base import BaseReduceRouter
from ...proto import gnes_pb2


class DocFillRouter(BaseReduceRouter):
    def apply(self, msg: 'gnes_pb2.Message', accum_msgs: List['gnes_pb2.Message'], *args, **kwargs):
        final_docs = []
        for idx, r in enumerate(msg.response.search.topk_results):
            # get result from all shards, some may return None, we only take the first non-None doc
            final_docs.append([m.response.search.topk_results[idx] for m in accum_msgs if
                               m.response.search.topk_results[idx].doc.WhichOneof('raw_data') is not None][0])

        # resort all doc result as the doc_weight has been applied
        final_docs = sorted(final_docs, key=lambda x: x.score, reverse=True)
        msg.response.search.ClearField('topk_results')
        msg.response.search.topk_results.extend(final_docs[:msg.response.search.top_k])

        super().apply(msg, accum_msgs)



import os
import unittest

from gnes.cli.parser import set_preprocessor_service_parser, _set_client_parser
from gnes.client.base import ZmqClient
from gnes.proto import gnes_pb2, RequestGenerator, blob2array
from gnes.service.preprocessor import PreprocessorService


class TestFFmpeg(unittest.TestCase):

    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        self.yml_path = os.path.join(self.dirname, 'yaml', 'preprocessor-ffmpeg.yml')
        self.yml_path_2 = os.path.join(self.dirname, 'yaml', 'preprocessor-ffmpeg2.yml')
        self.yml_path_3 = os.path.join(self.dirname, 'yaml', 'preprocessor-ffmpeg3.yml')
        self.video_path = os.path.join(self.dirname, 'videos')
        self.video_bytes = [open(os.path.join(self.video_path, _), 'rb').read()
                            for _ in os.listdir(self.video_path)]

    def test_video_preprocessor_service_empty(self):
        args = set_preprocessor_service_parser().parse_args([
            '--yaml_path', self.yml_path
        ])
        with PreprocessorService(args):
            pass

    def test_video_preprocessor_service_realdata(self):
        args = set_preprocessor_service_parser().parse_args([
            '--yaml_path', self.yml_path
        ])

        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in)
        ])

        with PreprocessorService(args), ZmqClient(c_args) as client:
            for req in RequestGenerator.index(self.video_bytes):
                msg = gnes_pb2.Message()
                msg.request.index.CopyFrom(req.index)
                client.send_message(msg)
                r = client.recv_message()
                for d in r.request.index.docs:
                    self.assertGreater(len(d.chunks), 0)
                    for _ in range(len(d.chunks)):
                        shape = blob2array(d.chunks[_].blob).shape
                        self.assertEqual(shape, (168, 192, 3))

    def test_video_cut_by_frame(self):
        args = set_preprocessor_service_parser().parse_args([
            '--yaml_path', self.yml_path_2,
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in)
        ])

        with PreprocessorService(args), ZmqClient(c_args) as client:
            for req in RequestGenerator.index(self.video_bytes):
                msg = gnes_pb2.Message()
                msg.request.index.CopyFrom(req.index)
                client.send_message(msg)
                r = client.recv_message()
                for d in r.request.index.docs:
                    self.assertGreater(len(d.chunks), 0)
                    for _ in range(len(d.chunks)-1):
                        shape = blob2array(d.chunks[_].blob).shape
                        self.assertEqual(shape, (30, 168, 192, 3))
                    shape = blob2array(d.chunks[-1].blob).shape
                    self.assertLessEqual(shape[0], 30)

    def test_video_cut_by_num(self):
        args = set_preprocessor_service_parser().parse_args([
            '--yaml_path', self.yml_path_3
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in)
        ])

        with PreprocessorService(args), ZmqClient(c_args) as client:
            for req in RequestGenerator.index(self.video_bytes):
                msg = gnes_pb2.Message()
                msg.request.index.CopyFrom(req.index)
                client.send_message(msg)
                r = client.recv_message()
                for d in r.request.index.docs:
                    self.assertEqual(len(d.chunks), 6)



import unittest

from gnes.cli.parser import set_frontend_parser


class TestParser(unittest.TestCase):
    def test_service_parser(self):
        args1 = set_frontend_parser().parse_args([])
        args2 = set_frontend_parser().parse_args([])
        self.assertNotEqual(args1.port_in, args2.port_in)
        self.assertNotEqual(args1.port_out, args2.port_out)



import os
import unittest

import numpy as np
from numpy.testing import assert_allclose

from gnes.encoder.base import PipelineEncoder
from gnes.encoder.numeric.pca import PCALocalEncoder
from gnes.encoder.numeric.pq import PQEncoder
from gnes.encoder.numeric.tf_pq import TFPQEncoder


class TestPCA(unittest.TestCase):
    def setUp(self):
        self.test_vecs = np.random.random([1000, 100]).astype('float32')
        dirname = os.path.dirname(__file__)
        self.lopq_yaml_np = os.path.join(dirname, 'yaml', 'lopq-encoder-2-np.yml')
        self.lopq_yaml_tf = os.path.join(dirname, 'yaml', 'lopq-encoder-2-tf.yml')
        self.lopq_yaml_np2 = os.path.join(dirname, 'yaml', 'lopq-encoder-3.yml')

    def test_pq_assert(self):
        self._test_pq_assert(PQEncoder)
        self._test_pq_assert(TFPQEncoder)

    def test_pq_tfpq_identity(self):
        def _test_pq_tfpq_identity(pq1, pq2):
            pq1.train(self.test_vecs)
            out1 = pq1.encode(self.test_vecs)
            pq2._copy_from(pq1)
            out2 = pq2.encode(self.test_vecs)
            assert_allclose(out1, out2)

        _test_pq_tfpq_identity(PQEncoder(10), TFPQEncoder(10))
        _test_pq_tfpq_identity(TFPQEncoder(10), PQEncoder(10))

    def _test_pq_assert(self, cls):
        self.assertRaises(AssertionError, cls, 100, 0)
        self.assertRaises(AssertionError, cls, 100, 256)

        pq = cls(8)
        self.assertRaises(AssertionError, pq.train, self.test_vecs)

        pq = cls(101)
        self.assertRaises(AssertionError, pq.train, self.test_vecs)

    def _simple_assert(self, out, num_bytes, num_clusters):
        self.assertEqual(bytes, type(out))
        self.assertEqual(self.test_vecs.shape[0] * num_bytes, len(out))
        self.assertTrue(np.all(np.frombuffer(out, np.uint8) <= num_clusters))

    def test_assert_pca(self):
        self.assertRaises(AssertionError, PCALocalEncoder, 8, 3)
        self.assertRaises(AssertionError, PCALocalEncoder, 2, 3)

        pca = PCALocalEncoder(100, 2)
        self.assertRaises(AssertionError, pca.train, self.test_vecs)

        pca = PCALocalEncoder(8, 2)
        self.assertRaises(AssertionError, pca.train, self.test_vecs[:7])

        pca.train(self.test_vecs)
        out = pca.encode(self.test_vecs)
        self.assertEqual(out.shape[1], 8)
        self.assertEqual(out.shape[0], self.test_vecs.shape[0])

    def test_train_pca(self):
        num_bytes = 8
        num_clusters = 11
        lopq = PipelineEncoder.load_yaml(self.lopq_yaml_np2)
        lopq.train(self.test_vecs)
        out = lopq.encode(self.test_vecs)
        self._simple_assert(out, num_bytes, num_clusters)

    # def test_train_pca_assert(self):
    #     # from PCA
    #     self.assertRaises(AssertionError, LOPQEncoder, num_bytes=100, pca_output_dim=20)
    #     # from PCA
    #     self.assertRaises(AssertionError, LOPQEncoder, num_bytes=7, pca_output_dim=20)
    #     # from LOPQ, cluster too large
    #     self.assertRaises(AssertionError, LOPQEncoder, num_bytes=4, pca_output_dim=20, cluster_per_byte=256)

    def test_encode_backend(self):
        num_bytes = 8
        lopq = PipelineEncoder.load_yaml(self.lopq_yaml_tf)
        lopq.train(self.test_vecs)
        out = lopq.encode(self.test_vecs)
        self._simple_assert(out, num_bytes, 255)

        lopq2 = PipelineEncoder.load_yaml(self.lopq_yaml_np)
        lopq2.train(self.test_vecs)
        out = lopq2.encode(self.test_vecs)
        self._simple_assert(out, num_bytes, 255)

        # copy from lopq
        lopq2._copy_from(lopq)
        out2 = lopq2.encode(self.test_vecs)
        self._simple_assert(out, num_bytes, 255)

        self.assertEqual(out, out2)

    def test_encode_batching(self):
        num_bytes = 8
        lopq = PipelineEncoder.load_yaml(self.lopq_yaml_tf)
        lopq.train(self.test_vecs)
        out = lopq.encode(self.test_vecs, batch_size=32)
        self._simple_assert(out, num_bytes, 255)
        out2 = lopq.encode(self.test_vecs, batch_size=64)
        self.assertEqual(out, out2)

    # def test_num_cluster(self):
    #     def _test_num_cluster(num_bytes, num_cluster, backend):
    #         lopq = LOPQEncoder(num_bytes,
    #                            cluster_per_byte=num_cluster,
    #                            pca_output_dim=20, pq_backend=backend)
    #         lopq.train(self.test_vecs)
    #         out = lopq.encode(self.test_vecs)
    #         self._simple_assert(out, num_bytes, num_cluster)
    #
    #     _test_num_cluster(10, 3, 'numpy')
    #     _test_num_cluster(10, 3, 'tensorflow')
    #     _test_num_cluster(10, 5, 'numpy')
    #     _test_num_cluster(10, 5, 'tensorflow')



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

import numpy as np

from ..base import BaseImageEncoder
from ...helper import batching


class BaseONNXImageEncoder(BaseImageEncoder):
    batch_size = 64

    def __init__(self, model_name: str,
                 model_dir: str,
                 use_cuda: bool = False,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.model_dir = model_dir
        self.model_name = model_name
        self._use_cuda = use_cuda

    def post_init(self):
        import onnxruntime as ort

        self.sess = ort.InferenceSession(self.model_dir + '/' + self.model_name)
        inputs_info = self.sess.get_inputs()

        if len(inputs_info) != 1:
            raise ValueError('Now only support encoder with one input')
        else:
            self.input_name = inputs_info[0].name
            self.input_shape = inputs_info[0].shape
            self.input_type = inputs_info[0].type
            self.batch_size = self.input_shape[0]

    @batching
    def encode(self, img: List['np.ndarray'], *args, **kwargs) -> np.ndarray:
        pad_batch = 0
        if len(img) != self.input_shape[0]:
            pad_batch = self.input_shape[0] - len(img)
            for _ in range(pad_batch):
                img.append(np.zeros_like(img[0]))

        img_ = np.array(img, dtype=np.float32).transpose(0, 3, 1, 2)
        if list(img_.shape) != self.input_shape:
            raise ValueError('Map size not match net, expect', self.input_shape, ',got', img_.shape)

        result_npy = self.sess.run(None, {self.input_name: img_})
        if pad_batch != 0:
            return result_npy[0][0:len(img)]
        else:
            return result_npy[0]



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import ctypes
import io
import random

import numpy as np
from PIL import Image

from ..base import TrainableBase, CompositionalTrainableBase
from ..proto import gnes_pb2, array2blob


class BasePreprocessor(TrainableBase):
    doc_type = gnes_pb2.Document.UNKNOWN

    def __init__(self, start_doc_id: int = 0,
                 random_doc_id: bool = True,
                 uniform_doc_weight: bool = True,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start_doc_id = start_doc_id
        self.random_doc_id = random_doc_id
        self.uniform_doc_weight = uniform_doc_weight

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        doc.doc_id = self.start_doc_id if not self.random_doc_id else random.randint(0, ctypes.c_uint(-1).value)
        doc.doc_type = self.doc_type
        if not doc.weight and self.uniform_doc_weight:
            doc.weight = 1.0
        self.start_doc_id += 1


class PipelinePreprocessor(CompositionalTrainableBase):
    def apply(self, doc: 'gnes_pb2.Document') -> None:
        if not self.component:
            raise NotImplementedError
        for be in self.component:
            be.apply(doc)

    def train(self, data, *args, **kwargs):
        if not self.component:
            raise NotImplementedError
        for idx, be in enumerate(self.component):
            be.train(data, *args, **kwargs)
            if idx + 1 < len(self.component):
                data = be.apply(data, *args, **kwargs)


class BaseUnaryPreprocessor(BasePreprocessor):
    is_trained = True

    def __init__(self, doc_type: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.doc_type = doc_type

    def apply(self, doc: 'gnes_pb2.Document'):
        super().apply(doc)
        c = doc.chunks.add()
        c.doc_id = doc.doc_id
        c.offset_1d = 0
        c.weight = 1.
        if doc.raw_bytes:
            self.raw_to_chunk(c, doc.raw_bytes)
        else:
            self.logger.error('bad document: "raw_bytes" is empty!')

    def raw_to_chunk(self, chunk: 'gnes_pb2.Chunk', raw_bytes: bytes):
        if self.doc_type == gnes_pb2.Document.TEXT:
            chunk.text = raw_bytes.decode()
        elif self.doc_type == gnes_pb2.Document.IMAGE:
            img = np.array(Image.open(io.BytesIO(raw_bytes)))
            chunk.blob.CopyFrom(array2blob(img))
        elif self.doc_type == gnes_pb2.Document.VIDEO:
            raise NotImplementedError
        else:
            raise NotImplementedError



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import importlib.util
import os
import sys

import grpc

from .base import BaseService as BS, MessageHandler
from ..proto import gnes_pb2


class GRPCService(BS):
    handler = MessageHandler(BS.handler)

    def post_init(self):
        self.channel = grpc.insecure_channel(
            '%s:%s' % (self.args.grpc_host, self.args.grpc_port),
            options=[('grpc.max_send_message_length', self.args.max_message_size * 1024 * 1024),
                     ('grpc.max_receive_message_length', self.args.max_message_size * 1024 * 1024)])

        foo = self.PathImport().add_modules(self.args.pb2_path, self.args.pb2_grpc_path)

        # build stub
        self.stub = getattr(foo, self.args.stub_name)(self.channel)

    def close(self):
        self.channel.close()
        super().close()

    @handler.register(NotImplementedError)
    def _handler_default(self, msg: 'gnes_pb2.Message'):
        yield getattr(self.stub, self.args.api_name)(msg)

    class PathImport:

        @staticmethod
        def get_module_name(absolute_path):
            module_name = os.path.basename(absolute_path)
            module_name = module_name.replace('.py', '')
            return module_name

        def add_modules(self, pb2_path, pb2_grpc_path):
            (module, spec) = self.path_import(pb2_path)
            sys.modules[spec.name] = module

            (module, spec) = self.path_import(pb2_grpc_path)
            sys.modules[spec.name] = module

            return module

        def path_import(self, absolute_path):
            module_name = self.get_module_name(absolute_path)
            spec = importlib.util.spec_from_file_location(module_name, absolute_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            sys.modules[spec.name] = module
            return module, spec



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.




import numpy as np

from ..base import BaseNumericEncoder
from ...helper import batching, train_required


class HashEncoder(BaseNumericEncoder):
    batch_size = 2048

    def __init__(self, num_bytes: int,
                 num_bits: int = 8,
                 num_idx: int = 3,
                 kmeans_clusters: int = 100,
                 method: str = 'product_uniform',
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert 1 <= num_bits <= 8, 'maximum 8 hash functions in a byte'
        self.num_bytes = num_bytes
        self.num_bits = num_bits
        self.num_idx = num_idx
        self.kmeans_clusters = kmeans_clusters
        self.method = method
        self.centroids = None
        self.x = None
        self.vec_dim = None
        self.hash_cores = None
        self.mean = None
        self.var = None

    def train(self, vecs: np.ndarray, *args, **kwargs):
        self.vec_dim = vecs.shape[1]
        self.centroids = [self.train_kmeans(vecs) for _ in range(self.num_idx)]
        self.centroids = np.reshape(
            self.centroids, [1, self.num_idx, self.kmeans_clusters, self.vec_dim]).astype(np.float32)

        if self.vec_dim % self.num_bytes != 0:
            raise ValueError('vec dim should be divided by x')
        self.x = int(self.vec_dim / self.num_bytes)
        self.mean = np.mean(vecs, axis=0)
        self.var = np.var(vecs, axis=0)
        self.hash_cores = [self.ran_gen() for _ in range(self.num_bytes)]
        self.proj = np.array([2 ** i for i in range(self.num_bits)]).astype(np.int32)

    def train_kmeans(self, vecs):
        import faiss
        kmeans_instance = faiss.Kmeans(self.vec_dim,
                                       self.kmeans_clusters,
                                       niter=10)
        if vecs.dtype != np.float32:
            vecs = vecs.astype(np.float32)
        kmeans_instance.train(vecs)
        centroids = kmeans_instance.centroids
        return centroids

    def pred_kmeans(self, vecs):
        vecs = np.reshape(vecs, [vecs.shape[0], 1, 1, vecs.shape[1]])
        dist = np.sum(np.square(vecs - self.centroids), -1)
        return np.argmax(-dist, axis=-1).astype(np.uint32)

    def ran_gen(self):
        self.logger.info('hash functions with %s' % self.method)
        if self.method == 'product_uniform':
            return np.random.uniform(-1, 1, size=(self.x, self.num_bits)
                                     ).astype(np.float32)
        elif self.method == 'uniform':
            return np.random.uniform(-1, 1, size=(self.vec_dim, self.num_bits)
                                     ).astype(np.float32)
        elif self.method == 'ortho_uniform':
            from scipy.stats import ortho_group
            m = ortho_group.rvs(dim=max(self.vec_dim, self.num_bits)
                                ).astype(np.float32)
            if self.vec_dim >= self.num_bits:
                return m[:, :self.num_bits]
            else:
                return m[:self.vec_dim, :]

    def hash(self, vecs):
        ret = []
        if self.method == 'product_uniform':
            vecs = np.reshape(vecs, [vecs.shape[0], self.num_bytes, self.x])
            for i in range(self.num_bytes):
                out = np.greater(np.matmul(vecs[:, i, :], self.hash_cores[i]), 0)
                ret.append(np.sum(out * self.proj, axis=1, keepdims=1))
            return np.concatenate(ret, axis=1).astype(np.uint32)
        elif self.method == 'uniform' or self.method == 'ortho_uniform':
            for i in range(self.num_bytes):
                out = np.greater(np.matmul(vecs, self.hash_cores[i]), 0)
                ret.append(np.sum(out * self.proj, axis=1, keepdims=1))
            return np.concatenate(ret, axis=1).astype(np.uint32)

    @train_required
    @batching
    def encode(self, vecs: np.ndarray, *args, **kwargs) -> np.ndarray:
        if vecs.shape[1] != self.vec_dim:
            raise ValueError('input dimension error')
        clusters = self.pred_kmeans(vecs)
        vecs = (vecs - self.mean) / self.var
        outcome = self.hash(vecs)
        return np.concatenate([clusters, outcome], axis=1)

    def _copy_from(self, x: 'HashEncoder') -> None:
        self.num_bytes = x.num_bytes
        self.num_bits = x.num_bits
        self.num_idx = x.num_idx
        self.kmeans_clusters = x.kmeans_clusters
        self.centroids = x.centroids
        self.method = x.method
        self.x = x.x
        self.vec_dim = x.vec_dim
        self.hash_cores = x.hash_cores
        self.mean = x.mean
        self.var = x.var
        self.is_trained = x.is_trained



import os
import unittest

import numpy as np

from gnes.indexer.vector.faiss import FaissIndexer


class TestEUIndexer(unittest.TestCase):
    def setUp(self):
        self.toy_query = np.random.random([1000, 20]).astype(np.float32)
        self.toy_label = np.random.randint(0, 1e9, [1000, 2]).tolist()
        self.add_query = np.random.random([1000, 20]).astype(np.float32)
        self.add_label = np.random.randint(0, 1e9, [1000, 2]).tolist()

        self.sub_query = self.toy_query[:10]
        dirname = os.path.dirname(__file__)
        self.dump_path = os.path.join(dirname, 'test_eu_faiss')

    def tearDown(self):
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)

    def test_add(self):
        fd = FaissIndexer(20, 'HNSW32', self.dump_path)
        fd.add(self.toy_label, self.toy_query, [1.] * len(self.toy_label))
        self.assertEqual(fd.size, self.toy_query.shape[0])
        fd.add(self.add_label, self.add_query, [1.] * len(self.add_label))
        self.assertEqual(fd.size,
                         self.toy_query.shape[0] + self.add_query.shape[0])

    def test_query(self):
        fd = FaissIndexer(20, 'HNSW32', self.dump_path)
        fd.add(self.toy_label, self.toy_query, [1.] * len(self.toy_label))
        ret = fd.query(self.sub_query, top_k=5)
        self.assertEqual(len(ret), self.sub_query.shape[0])
        self.assertEqual(len(ret[0]), 5)

    def test_dump_load(self):
        with FaissIndexer(20, 'HNSW32', self.dump_path) as tmp:
            tmp.add(self.toy_label, self.toy_query, [1.] * len(self.toy_label))
            tmp.dump()

        with FaissIndexer.load(tmp.dump_full_path) as fd:
            ret = fd.query(self.sub_query, top_k=2)
            self.assertEqual(len(ret), self.sub_query.shape[0])
            self.assertEqual(len(ret[0]), 2)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.




from typing import List

import numpy as np

from ..base import BaseTextEncoder
from ...helper import batching, pooling_np


class FlairEncoder(BaseTextEncoder):
    is_trained = True

    def __init__(self, model_name: str = 'multi-forward-fast',
                 pooling_strategy: str = 'REDUCE_MEAN', *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.model_name = model_name
        self.pooling_strategy = pooling_strategy

    def post_init(self):
        from flair.embeddings import FlairEmbeddings
        self._flair = FlairEmbeddings(self.model_name)

    @batching
    def encode(self, text: List[str], *args, **kwargs) -> np.ndarray:
        from flair.data import Sentence
        # tokenize text
        batch_tokens = [Sentence(sent) for sent in text]

        flair_encodes = self._flair.embed(batch_tokens)

        pooled_data = []
        for sentence in flair_encodes:
            _layer_data = np.stack([s.embedding.numpy() for s in sentence])
            _pooled = pooling_np(_layer_data, self.pooling_strategy)
            pooled_data.append(_pooled)
        return np.array(pooled_data, dtype=np.float32)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.



from typing import List

import numpy as np

from ..base import BaseAudioEncoder
from ...helper import batching


class MfccEncoder(BaseAudioEncoder):
    batch_size = 64

    def __init__(self, n_mfcc: int = 13, sample_rate: int = 16000, max_length: int = 100, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.n_mfcc = n_mfcc
        self.sample_rate = sample_rate
        self.max_length = max_length

    @batching
    def encode(self, data: List['np.array'], *args, **kwargs) -> np.ndarray:
        import librosa

        mfccs = [np.array(librosa.feature.mfcc(y=audio, sr=self.sample_rate, n_mfcc=self.n_mfcc).T)
                 for audio in data]

        mfccs = [np.concatenate((mf, np.zeros((self.max_length - mf.shape[0], self.n_mfcc), dtype=np.float32)), axis=0)
                 if mf.shape[0] < self.max_length else mf[:self.max_length] for mf in mfccs]
        mfccs = [mfcc.reshape((1, -1)) for mfcc in mfccs]
        mfccs = np.squeeze(np.array(mfccs), axis=1)
        return mfccs



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Generator

from ..base import BaseMapRouter
from ...helper import batch_iterator
from ...proto import gnes_pb2


class PublishRouter(BaseMapRouter):

    def __init__(self, num_part: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_part = num_part

    def apply(self, msg: 'gnes_pb2.Message', *args, **kwargs) -> Generator:
        msg.envelope.num_part.append(self.num_part)
        yield msg


class DocBatchRouter(BaseMapRouter):

    def apply(self, msg: 'gnes_pb2.Message', *args, **kwargs) -> Generator:
        if self.batch_size and self.batch_size > 0:
            batches = [b for b in batch_iterator(msg.request.index.docs, self.batch_size)]
            num_part = len(batches)
            for p_idx, b in enumerate(batches, start=1):
                _msg = gnes_pb2.Message()
                _msg.CopyFrom(msg)
                _msg.request.index.ClearField('docs')
                _msg.request.index.docs.extend(b)
                _msg.envelope.part_id = p_idx
                _msg.envelope.num_part.append(num_part)
                yield _msg



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from typing import List, Generator

from ..base import TrainableBase
from ..proto import gnes_pb2, merge_routes


class BaseRouter(TrainableBase):
    def apply(self, msg: 'gnes_pb2.Message', *args, **kwargs):
        pass


class BaseMapRouter(BaseRouter):
    def apply(self, msg: 'gnes_pb2.Message', *args, **kwargs) -> Generator:
        pass


class BaseReduceRouter(BaseRouter):
    def apply(self, msg: 'gnes_pb2.Message', accum_msgs: List['gnes_pb2.Message'], *args, **kwargs) -> None:
        merge_routes(msg, accum_msgs)
        if len(msg.envelope.num_part) > 1:
            msg.envelope.num_part.pop()
        else:
            self.logger.warning(
                'message envelope says num_part=%s, means no further message reducing. '
                'ignore this if you explicitly set "num_part" in RouterService' % msg.envelope.num_part)



import os
import unittest

import numpy as np

from gnes.encoder.base import PipelineEncoder
from gnes.encoder.numeric.hash import HashEncoder


class TestHash(unittest.TestCase):
    def setUp(self):
        self.num_bytes = 16
        self.num_bits = 8
        self.num_idx = 2
        self.kmeans_clusters = 10
        self.x = 1000
        self.y = 128
        self.test_data = np.random.random([self.x, self.y]).astype(np.float32)
        dirname = os.path.dirname(__file__)
        self.hash_yaml = os.path.join(dirname, 'yaml', 'hash-encoder.yml')

    def test_train_pred(self):
        m = HashEncoder(self.num_bytes, self.num_bits,
                        self.num_idx, self.kmeans_clusters)
        m.train(self.test_data)
        self.assertEquals(self.num_idx, m.centroids.shape[1])
        self.assertEquals(self.kmeans_clusters, m.centroids.shape[2])
        self.assertEqual(self.y, m.centroids.shape[3])

        self.assertEqual(self.num_bytes, len(m.hash_cores))

        out = m.encode(self.test_data)
        self.assertEqual(self.x, out.shape[0])
        self.assertEqual(self.num_idx+self.num_bytes, out.shape[1])
        self.assertEqual(np.uint32, out.dtype)

    def test_yaml_load(self):
        pca_hash = PipelineEncoder.load_yaml(self.hash_yaml)
        pca_hash.train(self.test_data)
        out = pca_hash.encode(self.test_data)
        self.assertEqual(self.x, out.shape[0])
        self.assertEqual(self.num_idx+self.num_bytes, out.shape[1])



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains common code shared by all inception models.

Usage of arg scope:
  with slim.arg_scope(inception_arg_scope()):
    logits, end_points = inception.inception_v3(images, num_classes,
                                                is_training=is_training)

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

slim = tf.contrib.slim


def inception_arg_scope(weight_decay=0.00004,
                        use_batch_norm=True,
                        batch_norm_decay=0.9997,
                        batch_norm_epsilon=0.001,
                        activation_fn=tf.nn.relu,
                        batch_norm_updates_collections=tf.GraphKeys.UPDATE_OPS,
                        batch_norm_scale=False):
    """Defines the default arg scope for inception models.

    Args:
      weight_decay: The weight decay to use for regularizing the model.
      use_batch_norm: "If `True`, batch_norm is applied after each convolution.
      batch_norm_decay: Decay for batch norm moving average.
      batch_norm_epsilon: Small float added to variance to avoid dividing by zero
        in batch norm.
      activation_fn: Activation function for conv2d.
      batch_norm_updates_collections: Collection for the update ops for
        batch norm.
      batch_norm_scale: If True, uses an explicit `gamma` multiplier to scale the
        activations in the batch normalization layer.

    Returns:
      An `arg_scope` to use for the inception models.
    """
    batch_norm_params = {
        # Decay for the moving averages.
        'decay': batch_norm_decay,
        # epsilon to prevent 0s in variance.
        'epsilon': batch_norm_epsilon,
        # collection containing update_ops.
        'updates_collections': batch_norm_updates_collections,
        # use fused batch norm if possible.
        'fused': None,
        'scale': batch_norm_scale,
    }
    if use_batch_norm:
        normalizer_fn = slim.batch_norm
        normalizer_params = batch_norm_params
    else:
        normalizer_fn = None
        normalizer_params = {}
    # Set weight_decay for weights in Conv and FC layers.
    with slim.arg_scope([slim.conv2d, slim.fully_connected],
                        weights_regularizer=slim.l2_regularizer(weight_decay)):
        with slim.arg_scope(
                [slim.conv2d],
                weights_initializer=slim.variance_scaling_initializer(),
                activation_fn=activation_fn,
                normalizer_fn=normalizer_fn,
                normalizer_params=normalizer_params) as sc:
            return sc



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.




from typing import List

import numpy as np

from ..base import BaseTextEncoder
from ...helper import batching, pooling_simple


class Word2VecEncoder(BaseTextEncoder):
    is_trained = True

    def __init__(self, model_dir: str,
                 skiprows: int = 1,
                 dimension: int = 300,
                 pooling_strategy: str = 'REDUCE_MEAN', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_dir = model_dir
        self.skiprows = skiprows
        self.pooling_strategy = pooling_strategy
        self.dimension = dimension

    def post_init(self):
        from ...helper import Tokenizer
        count = 0
        self.word2vec_df = {}
        with open(self.model_dir, 'r') as f:
            for line in f.readlines():
                line = line.strip().split(' ')
                if count < self.skiprows:
                    count += 1
                    continue
                if len(line) > self.dimension:
                    self.word2vec_df[line[0]] = np.array([float(i) for i in line[1:]], dtype=np.float32)

        self.empty = np.zeros([self.dimension], dtype=np.float32)
        self.cn_tokenizer = Tokenizer()

    @batching
    def encode(self, text: List[str], *args, **kwargs) -> np.ndarray:
        # tokenize text
        batch_tokens = [self.cn_tokenizer.tokenize(sent) for sent in text]
        pooled_data = []

        for tokens in batch_tokens:
            _layer_data = [self.word2vec_df.get(token, self.empty) for token in tokens]
            pooled_data.append(pooling_simple(_layer_data, self.pooling_strategy))

        return np.array(pooled_data).astype(np.float32)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import multiprocessing
import random
import threading
import time
import types
from contextlib import ExitStack
from enum import Enum
from typing import Tuple, List, Union, Type

import zmq
import zmq.decorators as zmqd
from termcolor import colored

from ..base import TrainableBase, T
from ..cli.parser import resolve_yaml_path
from ..helper import set_logger
from ..proto import gnes_pb2, add_route, send_message, recv_message


class BetterEnum(Enum):
    def __str__(self):
        return self.name

    @classmethod
    def from_string(cls, s):
        try:
            return cls[s]
        except KeyError:
            raise ValueError()


class ReduceOp(BetterEnum):
    CONCAT = 0
    ALWAYS_ONE = 1


class ParallelType(BetterEnum):
    PUSH_BLOCK = 0
    PUSH_NONBLOCK = 1
    PUB_BLOCK = 2
    PUB_NONBLOCK = 3

    @property
    def is_push(self):
        return self.value == 0 or self.value == 1

    @property
    def is_block(self):
        return self.value == 0 or self.value == 2


class SocketType(BetterEnum):
    PULL_BIND = 0
    PULL_CONNECT = 1
    PUSH_BIND = 2
    PUSH_CONNECT = 3
    SUB_BIND = 4
    SUB_CONNECT = 5
    PUB_BIND = 6
    PUB_CONNECT = 7
    PAIR_BIND = 8
    PAIR_CONNECT = 9

    @property
    def is_bind(self):
        return self.value % 2 == 0


class BlockMessage(Exception):
    pass


class ComponentNotLoad(Exception):
    pass


class ServiceError(Exception):
    pass


class EventLoopEnd(Exception):
    pass


def build_socket(ctx: 'zmq.Context', host: str, port: int, socket_type: 'SocketType', identity: 'str' = None) -> Tuple[
    'zmq.Socket', str]:
    sock = {
        SocketType.PULL_BIND: lambda: ctx.socket(zmq.PULL),
        SocketType.PULL_CONNECT: lambda: ctx.socket(zmq.PULL),
        SocketType.SUB_BIND: lambda: ctx.socket(zmq.SUB),
        SocketType.SUB_CONNECT: lambda: ctx.socket(zmq.SUB),
        SocketType.PUB_BIND: lambda: ctx.socket(zmq.PUB),
        SocketType.PUB_CONNECT: lambda: ctx.socket(zmq.PUB),
        SocketType.PUSH_BIND: lambda: ctx.socket(zmq.PUSH),
        SocketType.PUSH_CONNECT: lambda: ctx.socket(zmq.PUSH),
        SocketType.PAIR_BIND: lambda: ctx.socket(zmq.PAIR),
        SocketType.PAIR_CONNECT: lambda: ctx.socket(zmq.PAIR)
    }[socket_type]()

    if socket_type.is_bind:
        host = BaseService.default_host
        if port is None:
            sock.bind_to_random_port('tcp://%s' % host)
        else:
            sock.bind('tcp://%s:%d' % (host, port))
    else:
        if port is None:
            sock.connect(host)
        else:
            sock.connect('tcp://%s:%d' % (host, port))

    if socket_type in {SocketType.SUB_CONNECT, SocketType.SUB_BIND}:
        sock.setsockopt(zmq.SUBSCRIBE, identity.encode('ascii') if identity else b'')
        # sock.setsockopt(zmq.SUBSCRIBE, b'')

    return sock, sock.getsockopt_string(zmq.LAST_ENDPOINT)


class MessageHandler:
    def __init__(self, mh: 'MessageHandler' = None):
        self.routes = {k: v for k, v in mh.routes.items()} if mh else {}
        self.logger = set_logger(self.__class__.__name__)

    def register(self, msg_type: Union[List, Tuple, type]):
        def decorator(f):
            if isinstance(msg_type, list) or isinstance(msg_type, tuple):
                for m in msg_type:
                    self.routes[m] = f
            else:
                self.routes[msg_type] = f
            return f

        return decorator

    def serve(self, msg: 'gnes_pb2.Message'):
        def get_default_fn(m_type):
            self.logger.warning('cant find handler for message type: %s, fall back to the default handler' % m_type)
            f = self.routes.get(m_type, self.routes[NotImplementedError])
            return f

        if msg.WhichOneof('body'):
            body = getattr(msg, msg.WhichOneof('body'))
            if body.WhichOneof('body'):
                msg_type = type(getattr(body, body.WhichOneof('body')))
                if msg_type in self.routes:
                    self.logger.info('received a %r message' % msg_type.__name__)
                    fn = self.routes.get(msg_type)
                else:
                    fn = get_default_fn(msg_type)
            else:
                fn = get_default_fn(type(body))
        else:
            fn = get_default_fn(type(msg))
        return fn


class ConcurrentService(type):
    _dct = {}

    def __new__(cls, name, bases, dct):
        _cls = super().__new__(cls, name, bases, dct)
        ConcurrentService._dct.update({name: {'cls': cls,
                                              'name': name,
                                              'bases': bases,
                                              'dct': dct}})
        return _cls

    def __call__(cls, *args, **kwargs):
        # switch to the new backend
        _cls = {
            'thread': threading.Thread,
            'process': multiprocessing.Process
        }[args[0].parallel_backend]

        # rebuild the class according to mro
        for c in cls.mro()[-2::-1]:
            arg_cls = ConcurrentService._dct[c.__name__]['cls']
            arg_name = ConcurrentService._dct[c.__name__]['name']
            arg_dct = ConcurrentService._dct[c.__name__]['dct']
            _cls = super().__new__(arg_cls, arg_name, (_cls,), arg_dct)

        return type.__call__(_cls, *args, **kwargs)


class BaseService(metaclass=ConcurrentService):
    handler = MessageHandler()
    default_host = '0.0.0.0'

    def _get_event(self):
        if isinstance(self, threading.Thread):
            return threading.Event()
        elif isinstance(self, multiprocessing.Process):
            return multiprocessing.Event()
        else:
            raise NotImplementedError

    def __init__(self, args):
        super().__init__()
        self.args = args
        self.logger = set_logger(self.__class__.__name__, self.args.verbose)
        self.is_ready = self._get_event()
        self.is_event_loop = self._get_event()
        self.is_model_changed = self._get_event()
        self.is_handler_done = self._get_event()
        self._model = None
        self.identity = args.identity if 'identity' in args else None
        self.use_event_loop = True
        self.ctrl_addr = 'tcp://%s:%d' % (self.default_host, self.args.port_ctrl)

    def run(self):
        self._run()

    def _start_auto_dump(self):
        if self.args.dump_interval > 0 and not self.args.read_only:
            self._auto_dump_thread = threading.Thread(target=self._auto_dump)
            self._auto_dump_thread.setDaemon(1)
            self._auto_dump_thread.start()

    def _auto_dump(self):
        while self.is_event_loop.is_set():
            if self.is_model_changed.is_set():
                self.is_model_changed.clear()
                self.logger.info(
                    'auto-dumping the new change of the model every %ds...' % self.args.dump_interval)
                self.dump()
                time.sleep(self.args.dump_interval)
            else:
                time.sleep(1)

    def dump(self):
        if not self.args.read_only:
            if self._model:
                self.logger.info('dumping changes to the model...')
                self._model.dump()
                self.logger.info('dumping finished!')
        else:
            self.logger.warning('dumping is not allowed as "read_only" is set to true.')

    def message_handler(self, msg: 'gnes_pb2.Message', out_sck, ctrl_sck):
        try:
            fn = self.handler.serve(msg)
            if fn:
                add_route(msg.envelope, '%s:%s' % (self.__class__.__name__, self._model.__class__.__name__))
                self.logger.info(
                    'handling a message with route: %s' % '->'.join([r.service for r in msg.envelope.routes]))
                if msg.request and msg.request.WhichOneof('body') and \
                        type(getattr(msg.request, msg.request.WhichOneof('body'))) == gnes_pb2.Request.ControlRequest:
                    out_sock = ctrl_sck
                else:
                    out_sock = out_sck
                try:
                    # NOTE that msg is mutable object, it may be modified in fn()
                    ret = fn(self, msg)
                    if ret is None:
                        # assume 'msg' is modified inside fn()
                        send_message(out_sock, msg, timeout=self.args.timeout)
                    elif isinstance(ret, types.GeneratorType):
                        for r_msg in ret:
                            send_message(out_sock, r_msg, timeout=self.args.timeout)
                    else:
                        raise ServiceError('unknown return type from the handler: %s' % fn)

                    self.logger.info('handler %s is done' % fn.__name__)
                except BlockMessage:
                    pass
                except EventLoopEnd:
                    send_message(out_sock, msg, timeout=self.args.timeout)
                    raise EventLoopEnd
        except ServiceError as e:
            self.logger.error(e)

    @zmqd.context()
    def _run(self, ctx):
        ctx.setsockopt(zmq.LINGER, 0)
        self.logger.info('bind sockets...')
        in_sock, _ = build_socket(ctx, self.args.host_in, self.args.port_in, self.args.socket_in,
                                  getattr(self, 'identity', None))
        out_sock, _ = build_socket(ctx, self.args.host_out, self.args.port_out, self.args.socket_out,
                                   getattr(self, 'identity', None))
        ctrl_sock, ctrl_addr = build_socket(ctx, self.default_host, self.args.port_ctrl, SocketType.PAIR_BIND)

        self.logger.info(
            'input %s:%s\t output %s:%s\t control over %s' % (
                self.args.host_in, colored(self.args.port_in, 'yellow'),
                self.args.host_out, colored(self.args.port_out, 'yellow'),
                colored(ctrl_addr, 'yellow')))

        poller = zmq.Poller()
        poller.register(in_sock, zmq.POLLIN)
        poller.register(ctrl_sock, zmq.POLLIN)

        try:
            self.post_init()
            self.is_ready.set()
            self.is_event_loop.set()
            self._start_auto_dump()
            self.logger.info(colored('ready and listening', color='white', on_color='on_green'))
            while self.is_event_loop.is_set():
                pull_sock = None
                socks = dict(poller.poll())
                if socks.get(in_sock) == zmq.POLLIN:
                    pull_sock = in_sock
                elif socks.get(ctrl_sock) == zmq.POLLIN:
                    pull_sock = ctrl_sock
                else:
                    self.logger.error('received message from unknown socket: %s' % socks)
                if self.use_event_loop or pull_sock == ctrl_sock:
                    self.is_handler_done.clear()
                    msg = recv_message(pull_sock)
                    self.message_handler(msg, out_sock, ctrl_sock)
                    self.is_handler_done.set()
                else:
                    self.logger.warning(
                        'received a new message but since "use_event_loop=False" I will not handle it. '
                        'I will just block the thread until "is_handler_done" is set!')
                    self.is_handler_done.wait()
                    self.is_handler_done.clear()
                if self.args.dump_interval == 0:
                    self.dump()
        except EventLoopEnd:
            self.logger.info('break from the event loop')
        except ComponentNotLoad:
            self.logger.error('component can not be correctly loaded, terminated')
        finally:
            self.is_ready.set()
            self.is_event_loop.clear()
            in_sock.close()
            out_sock.close()
            ctrl_sock.close()
        self.logger.info('terminated')

    def post_init(self):
        pass

    def load_model(self, base_class: Type[TrainableBase]) -> T:
        try:
            return base_class.load_yaml(self.args.yaml_path)
        except FileNotFoundError:
            raise ComponentNotLoad

    @handler.register(NotImplementedError)
    def _handler_default(self, msg: 'gnes_pb2.Message'):
        raise NotImplementedError

    @handler.register(gnes_pb2.Request.ControlRequest)
    def _handler_control(self, msg: 'gnes_pb2.Message'):
        if msg.request.control.command == gnes_pb2.Request.ControlRequest.TERMINATE:
            self.is_event_loop.clear()
            msg.response.control.status = gnes_pb2.Response.SUCCESS
            raise EventLoopEnd
        elif msg.request.control.command == gnes_pb2.Request.ControlRequest.STATUS:
            msg.response.control.status = gnes_pb2.Response.ERROR
        else:
            raise ServiceError('dont know how to handle %s' % msg.request.control)

    def close(self):
        if self._model:
            self.dump()
            self._model.close()

        if self.is_event_loop.is_set():
            msg = gnes_pb2.Message()
            msg.request.control.command = gnes_pb2.Request.ControlRequest.TERMINATE
            return send_ctrl_message(self.ctrl_addr, msg, timeout=self.args.timeout)

    @property
    def status(self):
        msg = gnes_pb2.Message()
        msg.request.control.command = gnes_pb2.Request.ControlRequest.STATUS
        return send_ctrl_message(self.ctrl_addr, msg, timeout=self.args.timeout)

    def __enter__(self):
        self.start()
        self.is_ready.wait()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


def send_ctrl_message(address: str, msg: 'gnes_pb2.Message', timeout: int):
    # control message is short, set a timeout and ask for quick response
    with zmq.Context() as ctx:
        ctx.setsockopt(zmq.LINGER, 0)
        sock, _ = build_socket(ctx, address, None, SocketType.PAIR_CONNECT)
        send_message(sock, msg, timeout)
        r = recv_message(sock, timeout)
        sock.close()
        return r


class ServiceManager:
    def __init__(self, service_cls, args):
        self.logger = set_logger(self.__class__.__name__, args.verbose)
        self.services = []  # type: List['BaseService']
        if args.num_parallel > 1:
            from .router import RouterService
            _head_router = copy.deepcopy(args)
            _head_router.port_ctrl = self._get_random_port()
            port_out = self._get_random_port()
            _head_router.port_out = port_out

            _tail_router = copy.deepcopy(args)
            port_in = self._get_random_port()
            _tail_router.port_in = port_in
            _tail_router.port_ctrl = self._get_random_port()

            _tail_router.socket_in = SocketType.PULL_BIND

            if args.parallel_type.is_push:
                _head_router.socket_out = SocketType.PUSH_BIND
            else:
                _head_router.socket_out = SocketType.PUB_BIND
                _head_router.yaml_path = resolve_yaml_path(
                    '!PublishRouter {parameter: {num_part: %d}}' % args.num_parallel)

            if args.parallel_type.is_block:
                _tail_router.yaml_path = resolve_yaml_path('BaseReduceRouter')
                _tail_router.num_part = args.num_parallel

            self.services.append(RouterService(_head_router))
            self.services.append(RouterService(_tail_router))

            for _ in range(args.num_parallel):
                _args = copy.deepcopy(args)
                _args.port_in = port_out
                _args.port_out = port_in
                _args.port_ctrl = self._get_random_port()
                _args.socket_out = SocketType.PUSH_CONNECT
                if args.parallel_type.is_push:
                    _args.socket_in = SocketType.PULL_CONNECT
                else:
                    _args.socket_in = SocketType.SUB_CONNECT
                self.services.append(service_cls(_args))
            self.logger.info('num_parallel=%d, add a router with port_in=%d and a router with port_out=%d' % (
                args.num_parallel, _head_router.port_in, _tail_router.port_out))
        else:
            self.services.append(service_cls(args))

    @staticmethod
    def _get_random_port(min_port: int = 49152, max_port: int = 65536) -> int:
        return random.randrange(min_port, max_port)

    def __enter__(self):
        self.stack = ExitStack()
        for s in self.services:
            self.stack.enter_context(s)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stack.close()

    def join(self):
        for s in self.services:
            s.join()



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import inspect
import os
import pickle
import tempfile
import uuid
from functools import wraps
from typing import Dict, Any, Union, TextIO, TypeVar, Type, List, Callable

import ruamel.yaml.constructor

from ..helper import set_logger, profiling, yaml, parse_arg, load_contrib_module

__all__ = ['TrainableBase', 'CompositionalTrainableBase']

T = TypeVar('T', bound='TrainableBase')


def register_all_class(cls2file_map: Dict, module_name: str):
    import importlib
    for k, v in cls2file_map.items():
        try:
            getattr(importlib.import_module('gnes.%s.%s' % (module_name, v)), k)
        except ImportError:
            # print(e)
            pass
    load_contrib_module()


def import_class_by_str(name: str):
    def _import(module_name, class_name):
        import importlib

        cls2file = getattr(importlib.import_module('gnes.%s' % module_name), '_cls2file_map')
        if class_name in cls2file:
            return getattr(importlib.import_module('gnes.%s.%s' % (module_name, cls2file[class_name])), class_name)

    search_modules = ['encoder', 'indexer', 'preprocessor', 'router']

    for m in search_modules:
        r = _import(m, name)
        if r:
            return r
    else:
        raise ImportError('Can not locate any class with name: %s, misspelling?' % name)


class TrainableType(type):
    default_gnes_config = {
        'is_trained': False,
        'batch_size': None,
        'work_dir': os.environ.get('GNES_VOLUME', os.getcwd()),
        'name': None
    }

    def __new__(cls, *args, **kwargs):
        _cls = super().__new__(cls, *args, **kwargs)
        return cls.register_class(_cls)

    def __call__(cls, *args, **kwargs):
        # do _preload_package
        getattr(cls, '_pre_init', lambda *x: None)()

        if 'gnes_config' in kwargs:
            gnes_config = kwargs.pop('gnes_config')
        else:
            gnes_config = {}

        obj = type.__call__(cls, *args, **kwargs)

        # set attribute with priority
        # gnes_config in YAML > class attribute > default_gnes_config
        for k, v in TrainableType.default_gnes_config.items():
            if k in gnes_config:
                v = gnes_config[k]
            if not hasattr(obj, k):
                setattr(obj, k, v)

        getattr(obj, '_post_init_wrapper', lambda *x: None)()
        return obj

    @staticmethod
    def register_class(cls):
        # print('try to register class: %s' % cls.__name__)
        reg_cls_set = getattr(cls, '_registered_class', set())
        if cls.__name__ not in reg_cls_set:
            # print('reg class: %s' % cls.__name__)
            cls.__init__ = TrainableType._store_init_kwargs(cls.__init__)
            if os.environ.get('GNES_PROFILING', False):
                for f_name in ['train', 'encode', 'add', 'query']:
                    if getattr(cls, f_name, None):
                        setattr(cls, f_name, profiling(getattr(cls, f_name)))

            if getattr(cls, 'train', None):
                setattr(cls, 'train', TrainableType._as_train_func(getattr(cls, 'train')))

            reg_cls_set.add(cls.__name__)
            setattr(cls, '_registered_class', reg_cls_set)
            yaml.register_class(cls)
        return cls

    @staticmethod
    def _as_train_func(func):
        @wraps(func)
        def arg_wrapper(self, *args, **kwargs):
            if self.is_trained:
                self.logger.warning('"%s" has been trained already, '
                                    'training it again will override the previous training' % self.__class__.__name__)
            f = func(self, *args, **kwargs)
            self.is_trained = True
            return f

        return arg_wrapper

    @staticmethod
    def _store_init_kwargs(func):
        @wraps(func)
        def arg_wrapper(self, *args, **kwargs):
            taboo = {'self', 'args', 'kwargs'}
            taboo.update(TrainableType.default_gnes_config.keys())
            all_pars = inspect.signature(func).parameters
            tmp = {k: v.default for k, v in all_pars.items() if k not in taboo}
            tmp_list = [k for k in all_pars.keys() if k not in taboo]
            # set args by aligning tmp_list with arg values
            for k, v in zip(tmp_list, args):
                tmp[k] = v
            # set kwargs
            for k, v in kwargs.items():
                if k in tmp:
                    tmp[k] = v

            if self.store_args_kwargs:
                if args: tmp['args'] = args
                if kwargs: tmp['kwargs'] = {k: v for k, v in kwargs.items() if k not in taboo}

            if getattr(self, '_init_kwargs_dict', None):
                self._init_kwargs_dict.update(tmp)
            else:
                self._init_kwargs_dict = tmp
            f = func(self, *args, **kwargs)
            return f

        return arg_wrapper


class TrainableBase(metaclass=TrainableType):
    store_args_kwargs = False

    def __init__(self, *args, **kwargs):
        self.verbose = 'verbose' in kwargs and kwargs['verbose']
        self.logger = set_logger(self.__class__.__name__, self.verbose)
        self._post_init_vars = set()

    def _post_init_wrapper(self):
        if not getattr(self, 'name', None) and os.environ.get('GNES_WARN_UNNAMED_COMPONENT', '1') == '1':
            _id = str(uuid.uuid4()).split('-')[0]
            _name = '%s-%s' % (self.__class__.__name__, _id)
            self.logger.warning(
                'this object is not named ("- gnes_config: - name" is not found in YAML config), '
                'i will call it as "%s". '
                'naming the object is important especially when you need to '
                'serialize/deserialize/store/load the object.' % _name)
            setattr(self, 'name', _name)

        _before = set(list(self.__dict__.keys()))
        self.post_init()
        self._post_init_vars = {k for k in self.__dict__ if k not in _before}

    def post_init(self):
        pass

    @classmethod
    def pre_init(cls):
        pass

    @property
    def dump_full_path(self):
        return os.path.join(self.work_dir, '%s.bin' % self.name)

    @property
    def yaml_full_path(self):
        return os.path.join(self.work_dir, '%s.yml' % self.name)

    def __getstate__(self):
        d = dict(self.__dict__)
        del d['logger']
        if '_file_lock' in d:
            del d['_file_lock']
        for k in self._post_init_vars:
            del d[k]
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)
        self.logger = set_logger(self.__class__.__name__, self.verbose)
        try:
            self._post_init_wrapper()
        except ImportError:
            self.logger.info('ImportError is often caused by a missing component, '
                             'which often can be solved by "pip install" relevant package.')

    def train(self, *args, **kwargs):
        pass

    @profiling
    def dump(self, filename: str = None) -> None:
        f = filename or self.dump_full_path
        if not f:
            f = tempfile.NamedTemporaryFile('w', delete=False, dir=os.environ.get('GNES_VOLUME', None)).name
        with open(f, 'wb') as fp:
            pickle.dump(self, fp)
        self.logger.info('model is stored to %s' % f)

    @profiling
    def dump_yaml(self, filename: str = None) -> None:
        f = filename or self.yaml_full_path
        if not f:
            f = tempfile.NamedTemporaryFile('w', delete=False, dir=os.environ.get('GNES_VOLUME', None)).name
        with open(f, 'w') as fp:
            yaml.dump(self, fp)
        self.logger.info('model\'s yaml config is dump to %s' % f)

    @classmethod
    def load_yaml(cls: Type[T], filename: Union[str, TextIO]) -> T:
        if not filename: raise FileNotFoundError
        if isinstance(filename, str):
            with open(filename) as fp:
                return yaml.load(fp)
        else:
            with filename:
                return yaml.load(filename)

    @staticmethod
    @profiling
    def load(filename: str = None) -> T:
        if not filename: raise FileNotFoundError
        with open(filename, 'rb') as fp:
            return pickle.load(fp)

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @staticmethod
    def _get_tags_from_node(node):
        def node_recurse_generator(n):
            if n.tag.startswith('!'):
                yield n.tag.lstrip('!')
            for nn in n.value:
                if isinstance(nn, tuple):
                    for k in nn:
                        yield from node_recurse_generator(k)
                elif isinstance(nn, ruamel.yaml.nodes.Node):
                    yield from node_recurse_generator(nn)

        return list(set(list(node_recurse_generator(node))))

    @classmethod
    def to_yaml(cls, representer, data):
        tmp = data._dump_instance_to_yaml(data)
        return representer.represent_mapping('!' + cls.__name__, tmp)

    @classmethod
    def from_yaml(cls, constructor, node, stop_on_import_error=False):
        return cls._get_instance_from_yaml(constructor, node, stop_on_import_error)[0]

    @classmethod
    def _get_instance_from_yaml(cls, constructor, node, stop_on_import_error=False):
        try:
            for c in cls._get_tags_from_node(node):
                import_class_by_str(c)
        except ImportError as ex:
            if stop_on_import_error:
                raise RuntimeError('Cannot import module, pip install may required') from ex

        if node.tag in {'!PipelineEncoder', '!CompositionalTrainableBase'}:
            os.environ['GNES_WARN_UNNAMED_COMPONENT'] = '0'

        data = ruamel.yaml.constructor.SafeConstructor.construct_mapping(
            constructor, node, deep=True)

        dump_path = cls._get_dump_path_from_config(data.get('gnes_config', {}))
        load_from_dump = False
        if dump_path:
            obj = cls.load(dump_path)
            obj.logger.info('restore %s from %s' % (cls.__name__, dump_path))
            load_from_dump = True
        else:
            cls.init_from_yaml = True

            if cls.store_args_kwargs:
                p = data.get('parameter', {})  # type: Dict[str, Any]
                a = p.pop('args') if 'args' in p else ()
                k = p.pop('kwargs') if 'kwargs' in p else {}
                # maybe there are some hanging kwargs in "parameter"
                tmp_a = (cls._convert_env_var(v) for v in a)
                tmp_p = {kk: cls._convert_env_var(vv) for kk, vv in {**k, **p}.items()}
                obj = cls(*tmp_a, **tmp_p, gnes_config=data.get('gnes_config', {}))
            else:
                tmp_p = {kk: cls._convert_env_var(vv) for kk, vv in data.get('parameter', {}).items()}
                obj = cls(**tmp_p, gnes_config=data.get('gnes_config', {}))

            obj.logger.info('initialize %s from a yaml config' % cls.__name__)
            cls.init_from_yaml = False

        if node.tag in {'!PipelineEncoder', '!CompositionalTrainableBase'}:
            os.environ['GNES_WARN_UNNAMED_COMPONENT'] = '1'

        return obj, data, load_from_dump

    @staticmethod
    def _get_dump_path_from_config(gnes_config: Dict):
        if 'work_dir' in gnes_config and 'name' in gnes_config:
            dump_path = os.path.join(gnes_config['work_dir'], '%s.bin' % gnes_config['name'])
            if os.path.exists(dump_path):
                return dump_path

    @staticmethod
    def _convert_env_var(v):
        if isinstance(v, str):
            return parse_arg(os.path.expandvars(v))
        else:
            return v

    @staticmethod
    def _dump_instance_to_yaml(data):
        # note: we only dump non-default property for the sake of clarity
        p = {k: getattr(data, k) for k, v in TrainableType.default_gnes_config.items() if getattr(data, k) != v}
        a = {k: v for k, v in data._init_kwargs_dict.items() if k not in TrainableType.default_gnes_config}
        r = {}
        if a:
            r['parameter'] = a
        if p:
            r['gnes_config'] = p
        return r

    def _copy_from(self, x: 'TrainableBase') -> None:
        pass


class CompositionalTrainableBase(TrainableBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._component = None  # type: List[T]

    @property
    def component(self) -> Union[List[T], Dict[str, T]]:
        return self._component

    @property
    def is_pipeline(self):
        return isinstance(self.component, list)

    @component.setter
    def component(self, comps: Callable[[], Union[list, dict]]):
        if not callable(comps):
            raise TypeError('component must be a callable function that returns '
                            'a List[BaseEncoder]')
        if not getattr(self, 'init_from_yaml', False):
            self._component = comps()
        else:
            self.logger.info('component is omitted from construction, '
                             'as it is initialized from yaml config')

    def close(self):
        super().close()
        # pipeline
        if isinstance(self.component, list):
            for be in self.component:
                be.close()
        # no typology
        elif isinstance(self.component, dict):
            for be in self.component.values():
                be.close()
        elif self.component is None:
            pass
        else:
            raise TypeError('component must be dict or list, received %s' % type(self.component))

    def _copy_from(self, x: T):
        if isinstance(self.component, list):
            for be1, be2 in zip(self.component, x.component):
                be1._copy_from(be2)
        elif isinstance(self.component, dict):
            for k, v in self.component.items():
                v._copy_from(x.component[k])
        else:
            raise TypeError('component must be dict or list, received %s' % type(self.component))

    @classmethod
    def to_yaml(cls, representer, data):
        tmp = super()._dump_instance_to_yaml(data)
        tmp['component'] = data.component
        return representer.represent_mapping('!' + cls.__name__, tmp)

    @classmethod
    def from_yaml(cls, constructor, node):
        obj, data, from_dump = super()._get_instance_from_yaml(constructor, node)
        if not from_dump and 'component' in data:
            obj.component = lambda: data['component']
        return obj
